package org.apidesign.apifest08.test;

import org.apidesign.apifest08.currency.TimeRangeSpecificConvertor;
import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Currency;
import java.util.Date;
import junit.framework.TestCase;
import org.apidesign.apifest08.currency.ConversionRate;
import org.apidesign.apifest08.currency.Convertor;
import org.apidesign.apifest08.currency.Convertor.ConversionResult;
import org.apidesign.apifest08.currency.MonetaryAmount;
import org.apidesign.apifest08.currency.TableConvertor;
import org.apidesign.apifest08.currency.TimeRangeSpecificConvertor.TimeSpecificConversionRequest;

/** The exchange rates are not always the same. They are changing. However
 * as in order to predict the future, one needs to understand own past. That is
 * why it is important to know the exchange rate as it was at any time during
 * the past.
 * <p>
 * Today's quest is to enhance the convertor API to deal with dates.
 * One shall be able to convert a currency at any date. Each currencies rate shall
 * be associated with a range between two Date objects. In order
 * to keep compatibility with old API that knew nothing about dates, the
 * rates associated then are applicable "for eternity". Any use of existing
 * convert methods that do not accept a Date argument, uses the current
 * System.currentTimeMillis() as default date.
 */
public class Task4Test extends TestCase {
    public Task4Test(String testName) {
        super(testName);
    }

    @Override
    protected void setUp() throws Exception {
    }

    @Override
    protected void tearDown() throws Exception {
    }

    protected static final Currency CZK = Currency.getInstance( "CZK" );
    protected static final Currency SKK = Currency.getInstance( "SKK" );
    protected static final Currency USD = Currency.getInstance( "USD" );

    // Backward compatibly enhance your existing API to support following
    // usecases:
    //

    /** Takes a convertor with any rates associated and creates new convertor
     * that returns the same values as the old one for time between from to till.
     * Otherwise it returns no results. This is just a helper method that
     * shall call some real one in the API.
     * 
     * @param old existing convertor
     * @param from initial date (inclusive); null means since the Big Bang
     * @param till final date (exclusive); null means until the End of Universe
     * @return new convertor
     */
    public static Convertor limitTo( final Convertor old, final Date from, final Date till ) {
        return new TimeRangeSpecificConvertor( old, from, till );
    }
        
    private static final DateFormat DATE_FORMAT = new SimpleDateFormat( "yyyy-MM-dd HH:mm Z" );
    
    protected static Date parseGmtDate( final String string ) {
        final String zonedString = string + " GMT+0:00";
        try {
            return DATE_FORMAT.parse( zonedString );
        } catch ( final ParseException ex ) {
            throw new IllegalArgumentException( "Cannot parse " + zonedString, ex );
        }
    }

    public void testCompositionOfLimitedConvertors() throws Exception {
        final Date d1 = parseGmtDate( "2008-10-01 0:00" );
        final Date d2 = parseGmtDate( "2008-10-02 0:00" );
        final Date d3 = parseGmtDate( "2008-10-03 0:00" );
        
        final Convertor c = Task2Test.merge(
            limitTo(Task1Test.createCZKtoUSD(), d1, d2),
            limitTo(Task1Test.createSKKtoCZK(), d2, d3)
        );

        // convert $5 to CZK using c:
        final ConversionResult r1 = c.convert( new TimeSpecificConversionRequest( new MonetaryAmount( 5, USD ), CZK ) );
        final MonetaryAmount a1 = r1.getNetAmount();
        // cannot convert as no rate is applicable to current date
        assertNull( a1 );

        // convert $8 to CZK using c:
        final ConversionResult r2 = c.convert( new TimeSpecificConversionRequest( new MonetaryAmount( 8, USD ), CZK ) );
        final MonetaryAmount a2 = r2.getNetAmount();
        // cannot convert as no rate is applicable to current date
        assertNull( a2 );

        // convert 1003CZK to USD using c:
        final ConversionResult r3 = c.convert( new TimeSpecificConversionRequest( new MonetaryAmount( 1003, CZK ), USD ) );
        final MonetaryAmount a3 = r3.getNetAmount();
        // cannot convert as no rate is applicable to current date
        assertNull( a3 );

        // convert 16CZK using c:
        final ConversionResult r4 = c.convert( new TimeSpecificConversionRequest( new MonetaryAmount( 16, CZK ), USD ) );
        final MonetaryAmount a4 = r4.getNetAmount();
        // cannot convert as no rate is applicable to current date
        assertNull( a4 );

        // convert 500SKK to CZK using c:
        final ConversionResult r5 = c.convert( new TimeSpecificConversionRequest( new MonetaryAmount( 500, SKK ), CZK ) );
        final MonetaryAmount a5 = r5.getNetAmount();
        // cannot convert as no rate is applicable to current date
        assertNull( a5 );

        // convert $5 to CZK using c at 2008-10-01 6:00 GMT:
        final ConversionResult r6 = c.convert( new TimeSpecificConversionRequest( new MonetaryAmount( 5, USD ), CZK, parseGmtDate( "2008-10-01 6:00" ) ) );
        final MonetaryAmount a6 = r6.getNetAmount();
        // assertEquals("Result is 85 CZK");
        assertNotNull( a6 );
        assertEquals( 85.0, a6.getAmount().doubleValue() );
        assertEquals( CZK, a6.getCurrency() );

        // convert $8 to CZK using c at 2008-10-01 6:00 GMT:
        final ConversionResult r7 = c.convert( new TimeSpecificConversionRequest( new MonetaryAmount( 8, USD ), CZK, parseGmtDate( "2008-10-01 6:00" ) ) );
        final MonetaryAmount a7 = r7.getNetAmount();
        // assertEquals("Result is 136 CZK");
        assertNotNull( a7 );
        assertEquals( 136.0, a7.getAmount().doubleValue() );
        assertEquals( CZK, a7.getCurrency() );

        // convert 1003CZK to USD using c at 2008-10-01 6:00 GMT:
        final ConversionResult r8 = c.convert( new TimeSpecificConversionRequest( new MonetaryAmount( 1003, CZK ), USD, parseGmtDate( "2008-10-01 6:00" ) ) );
        final MonetaryAmount a8 = r8.getNetAmount();
        // assertEquals("Result is 59 USD");
        assertNotNull( a8 );
        assertEquals( 59.0, a8.getAmount().doubleValue() );
        assertEquals( USD, a8.getCurrency() );

        // convert 16CZK using c at 2008-10-02 9:00 GMT:
        final ConversionResult r9 = c.convert( new TimeSpecificConversionRequest( new MonetaryAmount( 16, CZK ), SKK, parseGmtDate( "2008-10-02 9:00" ) ) );
        final MonetaryAmount a9 = r9.getNetAmount();
        // assertEquals("Result is 20 SKK");
        assertNotNull( a9 );
        assertEquals( 20.0, a9.getAmount().doubleValue() );
        assertEquals( SKK, a9.getCurrency() );

        // convert 500SKK to CZK using c at 2008-10-02 9:00 GMT:
        final ConversionResult r10 = c.convert( new TimeSpecificConversionRequest( new MonetaryAmount( 500, SKK ), CZK, parseGmtDate( "2008-10-02 9:00" ) ) );
        final MonetaryAmount a10 = r10.getNetAmount();
        // assertEquals("Result is 400 CZK");
        assertNotNull( a10 );
        assertEquals( 400.0, a10.getAmount().doubleValue() );
        assertEquals( CZK, a10.getCurrency() );

        // convert 500SKK to CZK using c at 2008-10-01 6:00 GMT:
        final ConversionResult r11 = c.convert( new TimeSpecificConversionRequest( new MonetaryAmount( 500, SKK ), CZK, parseGmtDate( "2008-10-01 6:00" ) ) );
        final MonetaryAmount a11 = r11.getNetAmount();
        // cannot convert as no rate is applicable to current date
        assertNull( a11 );
    }

    /** Create convertor that understands two currencies, CZK and
     *  SKK. Make 100 SKK == 90 CZK.
     *
     * @return prepared convertor ready for converting SKK to CZK and CZK to SKK
     */
    public static Convertor createSKKtoCZK2() {
        final TableConvertor convertor = new TableConvertor();
        final MonetaryAmount amountInSKK = new MonetaryAmount( 100, SKK );
        final MonetaryAmount amountInCZK = new MonetaryAmount( 90, CZK );
        convertor.putIntoTable( new ConversionRate( amountInSKK, amountInCZK ) );
        convertor.putIntoTable( new ConversionRate( amountInCZK, amountInSKK ) );
        return new ContractImposingDelegatingConvertor( convertor ).test();
    }

    public void testDateConvetorWithTwoDifferentRates() throws Exception {
        final Date d1 = parseGmtDate( "2008-10-01 0:00" );
        final Date d2 = parseGmtDate( "2008-10-02 0:00" );
        final Date d3 = parseGmtDate( "2008-10-03 0:00" );

        final Convertor c = Task2Test.merge(
            limitTo(createSKKtoCZK2(), d1, d2),
            limitTo(Task1Test.createSKKtoCZK(), d2, d3)
        );

        // convert 500SKK to CZK using c at 2008-10-02 9:00 GMT:
        final ConversionResult r1 = c.convert( new TimeSpecificConversionRequest( new MonetaryAmount( 500, SKK ), CZK, parseGmtDate( "2008-10-02 9:00" ) ) );
        final MonetaryAmount a1 = r1.getNetAmount();
        // assertEquals("Result is 400 CZK");
        assertNotNull( a1 );
        assertEquals( 400.0, a1.getAmount().doubleValue() );
        assertEquals( CZK, a1.getCurrency() );

        // convert 500SKK to CZK using c at 2008-10-01 6:00 GMT:
        final ConversionResult r2 = c.convert( new TimeSpecificConversionRequest( new MonetaryAmount( 500, SKK ), CZK, parseGmtDate( "2008-10-01 6:00" ) ) );
        final MonetaryAmount a2 = r2.getNetAmount();
        // assertEquals("Result is 450 CZK");
        assertNotNull( a2 );
        assertEquals( 450.0, a2.getAmount().doubleValue() );
        assertEquals( CZK, a2.getCurrency() );        
    }

}
