package org.apidesign.apifest08.test;

import java.util.Date;
import java.util.GregorianCalendar;
import junit.framework.TestCase;
import org.apidesign.apifest08.currency.Convertor;
import org.apidesign.apifest08.currency.ConvertorFactory;

/** The exchange rates are not always the same. They are changing. However
 * as in order to predict the future, one needs to understand own past. That is
 * why it is important to know the exchange rate as it was at any time during
 * the past.
 * <p>
 * Today's quest is to enhance the convertor API to deal with dates.
 * One shall be able to convert a currency at any date. Each currencies rate shall
 * be associated with a range between two Date objects. In order
 * to keep compatibility with old API that knew nothing about dates, the
 * rates associated then are applicable "for eternity". Any use of existing
 * convert methods that do not accept a Date argument, uses the current
 * System.currentTimeMillis() as default date.
 */
public class Task4Test extends TestCase {
    public Task4Test(String testName) {
        super(testName);
    }

    @Override
    protected void setUp() throws Exception {
    }

    @Override
    protected void tearDown() throws Exception {
    }

    // Backward compatibly enhance your existing API to support following
    // usecases:
    //

    /** Takes a convertor with any rates associated and creates new convertor
     * that returns the same values as the old one for time between from to till.
     * Otherwise it returns no results. This is just a helper method that
     * shall call some real one in the API.
     * 
     * @param old existing convertor
     * @param from initial date (inclusive)
     * @param till final date (exclusive)
     * @return new convertor
     */
    public static Convertor limitTo(Convertor old, Date from, Date till) {
        return ConvertorFactory.newInstance().limitConvertor(old, from, till);
    }


    public void testCompositionOfLimitedConvertors() throws Exception {
        Date d1 = (new GregorianCalendar(2008, 9, 1, 0, 0)).getTime(); // 2008-10-01 0:00 GMT
        Date d2 = (new GregorianCalendar(2008, 9, 2, 0, 0)).getTime(); // 2008-10-02 0:00 GMT
        Date d3 = (new GregorianCalendar(2008, 9, 3, 0, 0)).getTime(); // 2008-10-03 0:00 GMT
        
        Convertor c = Task2Test.merge(
            limitTo(Task1Test.createCZKtoUSD(), d1, d2),
            limitTo(Task1Test.createSKKtoCZK(), d2, d3)
        );

        // convert $5 to CZK using c:
        // cannot convert as no rate is applicable to current date
        try {
            c.convert("USD", "CZK", 5);
            fail();
        } catch (IllegalArgumentException e) {
            //ok
        }

        // convert $8 to CZK using c:
        // cannot convert as no rate is applicable to current date
        try {
            c.convert("USD", "CZK", 8);
            fail();
        } catch (IllegalArgumentException e) {
            //ok
        }

        // convert 1003CZK to USD using c:
        // cannot convert as no rate is applicable to current date
        try {
            c.convert("CZK", "USD", 1003);
            fail();
        } catch (IllegalArgumentException e) {
            //ok
        }

        // convert 16CZK using c:
        // cannot convert as no rate is applicable to current date
        try {
            c.convert("CZK", "USD", 16);
            fail();
        } catch (IllegalArgumentException e) {
            //ok
        }

        // convert 500SKK to CZK using c:
        // cannot convert as no rate is applicable to current date
        try {
            c.convert("SKK", "CZK", 500);
            fail();
        } catch (IllegalArgumentException e) {
            //ok
        }

        // convert $5 to CZK using c at 2008-10-01 6:00 GMT:
        // assertEquals("Result is 85 CZK");
        assertEquals("Result is 85 CZK", 85.0, c.convert("USD", "CZK", 5, (new GregorianCalendar(2008,9,1,6,0)).getTime()));

        // convert $8 to CZK using c at 2008-10-01 6:00 GMT:
        // assertEquals("Result is 136 CZK");
        assertEquals("Result is 136 CZK", 136.0, c.convert("USD", "CZK", 8, (new GregorianCalendar(2008,9,1,6,0)).getTime()));

        // convert 1003CZK to USD using c at 2008-10-01 6:00 GMT:
        // assertEquals("Result is 59 USD");
        assertEquals("Result is 59 USD", 59.0, c.convert("CZK", "USD", 1003.0, (new GregorianCalendar(2008, 9, 1, 6, 0)).getTime()));

        // convert 16CZK using c at 2008-10-02 9:00 GMT:
        // assertEquals("Result is 20 SKK");
        assertEquals("Result is 20 SKK", 20.0, c.convert("CZK", "SKK", 16.0, (new GregorianCalendar(2008, 9, 2, 9, 0)).getTime()));

        // convert 500SKK to CZK using c at 2008-10-02 9:00 GMT:
        // assertEquals("Result is 400 CZK");
        assertEquals("Result is 400 CZK", 400.0, c.convert("SKK", "CZK", 500.0, (new GregorianCalendar(2008, 9, 2, 9, 0)).getTime()));

        // convert 500SKK to CZK using c at 2008-10-01 6:00 GMT:
        // cannot convert as no rate is applicable to current date
        try {
            c.convert("SKK", "CZK", 500.0, (new GregorianCalendar(2008, 9, 1, 6, 0)).getTime());
            fail();
        } catch (IllegalArgumentException e) {
            //ok
        }
    }

    /** Create convertor that understands two currencies, CZK and
     *  SKK. Make 100 SKK == 90 CZK.
     *
     * @return prepared convertor ready for converting SKK to CZK and CZK to SKK
     */
    public static Convertor createSKKtoCZK2() {
        return ConvertorFactory.newInstance().createConvertor("SKK", "CZK", 100, 90);
    }

    public void testDateConvetorWithTwoDifferentRates() throws Exception {
        Date d1 = (new GregorianCalendar(2008,9,1,0,0)).getTime(); // 2008-10-01 0:00 GMT
        Date d2 = (new GregorianCalendar(2008,9,2,0,0)).getTime(); // 2008-10-02 0:00 GMT
        Date d3 = (new GregorianCalendar(2008,9,6,0,0)).getTime(); // 2008-10-03 0:00 GMT

        Convertor c = Task2Test.merge(
            limitTo(createSKKtoCZK2(), d1, d2),
            limitTo(Task1Test.createSKKtoCZK(), d2, d3)
        );

        // convert 500SKK to CZK using c at 2008-10-02 9:00 GMT:
        // assertEquals("Result is 400 CZK");
        assertEquals("Result is 400 CZK", 400.0, c.convert("SKK", "CZK", 500.0, (new GregorianCalendar(2008, 9, 2, 9, 0)).getTime()));

        // convert 500SKK to CZK using c at 2008-10-01 6:00 GMT:
        // assertEquals("Result is 450 CZK");
        assertEquals("Result is 450 CZK", 450.0, c.convert("SKK", "CZK", 500.0, (new GregorianCalendar(2008, 9, 1, 6, 0)).getTime()));
    }

}
